import os
from collections import OrderedDict

from qgis.PyQt.QtCore import QCoreApplication
from qgis.core import (
    QgsProcessingAlgorithm,
    QgsProcessingException,
    QgsProcessingParameterFile,
    QgsProcessingParameterFileDestination,
    QgsSettings,
)
from ThreeDiToolbox.processing.results_analysis.utils import gridadmin2geopackage, load_computational_layers


class ThreeDiConvertToGpkgAlgorithm(QgsProcessingAlgorithm):
    """Convert gridadmin.h5 to GeoPackage with vector layers and add subset of those layers to the map canvas."""

    INPUT = "INPUT"
    OUTPUT = "OUTPUT"
    LAYERS_TO_ADD = OrderedDict()

    def flags(self):
        return super().flags() | QgsProcessingAlgorithm.FlagNoThreading

    def tr(self, string):
        return QCoreApplication.translate("Processing", string)

    def createInstance(self):
        return ThreeDiConvertToGpkgAlgorithm()

    def name(self):
        return "threedi_convert_gridadmin_to_gpkg"

    def displayName(self):
        return self.tr("Computational grid from gridadmin.h5 file")

    def group(self):
        return self.tr("Computational Grid")

    def groupId(self):
        return "computational_grid"

    def shortHelpString(self):
        return self.tr("Create computational grid from gridadmin.h5 file")

    def initAlgorithm(self, config=None):
        s = QgsSettings()
        last_input_dir = s.value("threedi-results-analysis/gridadmin_to_gpkg/last_input", None)
        last_output_gpkg = s.value("threedi-results-analysis/gridadmin_to_gpkg/last_output_gpkg", None)

        self.addParameter(
            QgsProcessingParameterFile(
                self.INPUT,
                self.tr("Folder containing input gridadmin.h5"),
                behavior=QgsProcessingParameterFile.Folder,
                defaultValue=last_input_dir,
            )
        )

        self.addParameter(
            QgsProcessingParameterFileDestination(
                self.OUTPUT, self.tr("Output GeoPackage file"), fileFilter="*.gpkg", defaultValue=last_output_gpkg
            )
        )

    def processAlgorithm(self, parameters, context, feedback):
        gridadmin_folder = self.parameterAsString(parameters, self.INPUT, context)
        if gridadmin_folder is None:
            raise QgsProcessingException(self.invalidSourceError(parameters, self.INPUT))
        input_gridadmin = os.path.join(gridadmin_folder, "gridadmin.h5")

        gpkg_path = self.parameterAsFileOutput(parameters, self.OUTPUT, context)
        if gpkg_path is None:
            raise QgsProcessingException(self.invalidSourceError(parameters, self.OUTPUT))
        if gpkg_path.endswith(".file"):
            gpkg_path = gpkg_path.rsplit(".", 1)[0] + ".gpkg"

        s = QgsSettings()
        s.setValue("threedi-results-analysis/gridadmin_to_gpkg/last_input", gridadmin_folder)
        s.setValue("threedi-results-analysis/gridadmin_to_gpkg/last_output_gpkg", gpkg_path)

        gpkg_layers = gridadmin2geopackage(input_gridadmin, gpkg_path, context, feedback)
        self.LAYERS_TO_ADD.update(gpkg_layers)

        return {}

    def postProcessAlgorithm(self, context, feedback):
        project = context.project()
        load_computational_layers(self.LAYERS_TO_ADD, project)
        self.LAYERS_TO_ADD.clear()
        return {}
